char a;
signed char as;
unsigned char au;
short b;
unsigned short bu;
int c;
unsigned int cu;
long d;
unsigned long du;
long long e;
unsigned long long eu;
void *f;
_Bool g;
float h;
double i;
long double j;
enum {k} l;
